package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.DistributeSshKeypair;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.util.lib.BestNodeLoginFinder;
import be.iminds.ilabt.jfed.util.library.SshProxySocketImplFactory;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.transport.TransportException;
import net.schmizz.sshj.transport.verification.PromiscuousVerifier;
import net.schmizz.sshj.userauth.UserAuthException;
import net.schmizz.sshj.userauth.keyprovider.KeyPairWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.net.SocketFactory;
import java.io.IOException;
import java.security.KeyPair;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

public abstract class JobWithSshConnectionManager<T> extends AbstractJob<T> {
    private static final Logger LOG = LoggerFactory.getLogger(JobWithSshConnectionManager.class);

    @Nonnull protected final JobStateFactory jobStateFactory;
    @Nonnull protected final ProxySocketFactoryProvider proxySocketFactoryProvider;
    @Nonnull protected final GeniUserProvider geniUserProvider;

    public JobWithSshConnectionManager(@Nonnull String name,
                                       @Nonnull Experiment experiment,
                                       @Nonnull HighLevelTaskFactory hltf,
                                       @Nonnull TaskThread tt,
                                       @Nonnull JobStateFactory jobStateFactory,
                                       @Nonnull ProxySocketFactoryProvider proxySocketFactoryProvider,
                                       @Nonnull GeniUserProvider geniUserProvider) {
        super(name, experiment, hltf, tt);
        this.jobStateFactory = jobStateFactory;
        this.proxySocketFactoryProvider = proxySocketFactoryProvider;
        this.geniUserProvider = geniUserProvider;
    }


    /**
     *
     * @param newState
     * @return success
     * @throws JFedException
     * @throws InterruptedException
     */
    protected boolean setAndRunSshUsingState(@Nonnull State newState) throws JFedException, InterruptedException {
        boolean sawException = true;
        boolean success = true;
        try {
            super.setAndRunState(newState);
            sawException = false;
        } catch (Throwable t) {
            sawException = true;
            success = false;
            throw t;
        } finally {
            if (sawException || newState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                success = false;
                LOG.warn("Something went wrong in "+newState.getName()+" ("+newState.getStatus()+") (ex="+sawException+") -> closing all SSH connections");
                closeAllSshConnections();
                //a bit dirty, but this works (and shouldn't mask exceptions)
                if (!sawException) {
                    //noinspection ThrowFromFinallyBlock
                    throw new RuntimeException("JobWithSshConnectionManager encountered a failure in: " + newState.getName());
                }
            }
        }
        return success;
    }



    private KeyPair loginKeyPair;
    private String loginUsername;

    @Nonnull
    public KeyPair getKeyPair() {
        if (this.loginKeyPair == null) {
            //find the keypair and username to login. This is either the distributed key, or if that is not present, just the user's key.
            assert experiment.getKeypairs() != null;
            loginUsername = null;
            KeyPair loginKeyPair = null;

            for (Map.Entry<DistributeSshKeypair, KeyPair> e : experiment.getKeypairs().entrySet()) {
                if(e.getValue() != null){
                    //distributedSshKeysUsername may be null, but that's ok, that just means the logged in users account needs to be used.
                    loginUsername = e.getKey().getUser() == null ? null : e.getKey().getUser().getResourceName();

                    loginKeyPair = e.getValue();
                    break;
                }
            }

            if(loginKeyPair == null) {
                loginKeyPair = new KeyPair(geniUserProvider.getLoggedInGeniUser().getPublicKey(),
                        geniUserProvider.getLoggedInGeniUser().getPrivateKey());
                loginUsername = geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();
            }

            this.loginKeyPair = loginKeyPair;
        }

        return this.loginKeyPair;
    }

    @Nonnull
    public String getLoginUsername() {
        if (this.loginUsername == null || this.loginKeyPair == null) {
            getKeyPair();
            if (loginUsername == null)
                throw new IllegalStateException("Did not find login username");
        }
        return loginUsername;
    }

    private BestNodeLoginFinder bestNodeLoginFinder;
    @Nullable
    public BestNodeLoginFinder getBestNodeLoginFinder() {
        if (bestNodeLoginFinder == null) {
            String distributedSshKeysUsername = getLoginUsername();

            ManifestRspecSource manifestRspecSource = experiment.getSlice().getManifestRspec();
            bestNodeLoginFinder = manifestRspecSource == null ||
                    manifestRspecSource.getStringRspec() == null ? null :
                    new BestNodeLoginFinder(manifestRspecSource.getStringRspec(),
                            distributedSshKeysUsername,
                            this.geniUserProvider.getLoggedInGeniUser(),
                            new BestNodeLoginFinder.Feedback() {
                                @Override
                                public void info(String s) {
                                    LOG.info("JobWithSshConnectionManager.BestNodeLoginFinder: " + s);
                                }

                                @Override
                                public void error(String s) {
                                    LOG.error("JobWithSshConnectionManager.BestNodeLoginFinder: " + s);
                                }

                                @Override
                                public void debug(String s) {
                                    LOG.debug("JobWithSshConnectionManager.BestNodeLoginFinder: " + s);
                                }
                            });
        }

        return bestNodeLoginFinder;
    }

    public static class SshException extends Exception {
        public SshException() { }
        public SshException(String s) { super(s); }
        public SshException(String s, Throwable throwable) { super(s, throwable); }
        public SshException(Throwable throwable) { super(throwable); }
        public SshException(String s, Throwable throwable, boolean b, boolean b1) {
            super(s, throwable, b, b1);
        }
    }

    private Map<String, SSHClient> sshClientByUniqueNodeId = new ConcurrentHashMap<>();

    /**
     * Will return a connected SSH connection. This can return an existing connection,
     * and will create a new one if needed.
     *
     * @param node node to which a connection is required
     * @param stateSlice may be null. If not, setAndRunState might be run on it to setup the connection
     * @return a connected SSHConnection. (never returns null)
     * @throws JFedException
     * @throws InterruptedException
     * @throws SshException if anything went wrong seting up the connection, or is wrong with the existing connection
     */
    @Nonnull
    public synchronized SSHClient getSSHClient(@Nonnull RspecNode node, @Nullable StateSlice stateSlice)
            throws JFedException, InterruptedException, SshException {
        SSHClient res = sshClientByUniqueNodeId.get(node.getUniqueId());
        if (res != null) {
            if (!res.isConnected()) {
                throw new SshException("Existing SSH connection is not connected (for node "+node.getUniqueId()+")");
//                sshClientByUniqueNodeId.remove(node.getUniqueId());
            } else {
                return res;
            }
        }

        final SetupSshConnectionState setupSshConnectionState = new SetupSshConnectionState(node);
        if (stateSlice == null) {
            setupSshConnectionState.executeState(this);
        } else {
            stateSlice.setAndRunState(setupSshConnectionState);
            if (setupSshConnectionState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                throw new SshException("Failed to setup SSH connection ("+setupSshConnectionState.getStatus()+") " +
                        "to node "+node.getClientId());
            }
        }
        res = setupSshConnectionState.getSsh();
        if (res == null) {
            throw new SshException("Failed to setup SSH connection (sshClient=null) to node "+node.getUniqueId());
        }
        sshClientByUniqueNodeId.put(node.getUniqueId(), res);

        if (!res.isConnected()) {
            throw new SshException("SSH connection is not connected with node "+node.getUniqueId());
        }
        return res;
    }

    /**
     * Possibly close the SSH connection. Call this if you don't need a connection anymore.
     *
     * @param node
     */
    public synchronized void closeSSH(@Nonnull RspecNode node) {
        //TODO: allow close only if needed

//        SSHClient ssh = sshClientByUniqueNodeId.remove(node.getUniqueId());
//        if (ssh == null)
//            return;
//
//        try {
//            ssh.close();
//        } catch (IOException ignored) {
//        }
    }

    public synchronized void closeAllSshConnections() {
        LOG.debug("closeAllSshConnections() called");
        synchronized (sshClientByUniqueNodeId) {
            for (SSHClient ssh : sshClientByUniqueNodeId.values()) {
                try {
                    if (ssh != null)
                        ssh.close();
                } catch (IOException ignored) {
                    //probably, connection is already closed. We don't really need to worry about this.
                    LOG.debug("Will ignore error while closing SSH connection: " + ignored.getMessage());
                }
            }
            //make sure it isn't reused
            sshClientByUniqueNodeId.clear();
        }
    }


    ////////////////////////////////////////////////////////

    private static final int SETUP_SSH_CONNECTION_RETRIES = 3;

    public class SetupSshConnectionState extends State {

        @Nonnull private final RspecNode node;
        private SSHClient ssh;

        protected SetupSshConnectionState(@Nonnull RspecNode node) {
            super("Setting up SSH Connection to '" + node.getClientId() + "'");
            this.node = node;
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            try {
                ssh = setupNewSSHClient(node);
            } catch (SshSetupException e) {
                return ExperimentTaskStatus.FAILED;
            }

            return ExperimentTaskStatus.SUCCESS;

        }

        public SSHClient getSsh() {
            return ssh;
        }
    }

    ////////////////////////////////////////////////////////


    public static class SshSetupException extends Exception {
        public SshSetupException() { }
        public SshSetupException(String message) { super(message); }
        public SshSetupException(String message, Throwable cause) { super(message, cause); }
        public SshSetupException(Throwable cause) { super(cause); }
        public SshSetupException(String message, Throwable cause, boolean enableSuppression, boolean writableStackTrace) {
            super(message, cause, enableSuppression, writableStackTrace);
        }
    }

    @Nonnull
    public SSHClient setupNewSSHClient(@Nonnull RspecNode node) throws SshSetupException {
        BestNodeLoginFinder loginFinder = getBestNodeLoginFinder();
        if (loginFinder == null) {
            LOG.warn("loginFinder is null! Will not be able to setup SSH connection.");
        }

        //get login-service associated to the user we are distributing the key to
        BasicStringRspec.LoginService loginService =
                loginFinder == null ?
                        null :
                        loginFinder.findBestLogin(node.getUniqueId());

        if (loginService == null) {
            LOG.warn("Cannot setup connection to {} as no login-service is available", node.getClientId());
            throw new SshSetupException("Cannot setup connection to "+node.getClientId()+" as no login-service is available");
//            return ExperimentTaskStatus.FAILED;
        }

        SSHClient ssh = new SSHClient();
        ssh.setTimeout(SshProxySocketImplFactory.KEX_TIMEOUT_MS);
        ssh.setConnectTimeout(SshProxySocketImplFactory.CONNECT_TIMEOUT_MS);
        ssh.addHostKeyVerifier(new PromiscuousVerifier());

        try {
            SocketFactory socketFactory = proxySocketFactoryProvider.createProxySocketFactoryForSsh(loginService.getHostname(), loginService.getPort());
            if (socketFactory != null) {
                LOG.debug("Using proxy for SSH access");
                ssh.setSocketFactory(socketFactory);
            } else {
                LOG.debug("No proxy needed for SSH access");
            }
        } catch (IOException ex) {
            LOG.error("Error while creating ProxySocketFactory. Trying without now.", ex);
        }

        int connectTries = 0;
        boolean connected = false;
        Exception lastConnectEx = null;

        while(!connected && connectTries < SETUP_SSH_CONNECTION_RETRIES){
            connectTries++;
            try {
                ssh.connect(loginService.getHostname(), loginService.getPort());
                connected = true;
            } catch (IOException e) {
                lastConnectEx = e;
                String message = "Error while connecting to " + loginService.getHostname() + ": " + e.getMessage();
                if(connectTries < SETUP_SSH_CONNECTION_RETRIES)
                    message += ". Retrying ...";
                updateMessage(message);
                LOG.error("Error while connecting to {}: {}", loginService.getHostname(), e.getMessage(), e);
            }
        }

        if(!connected){
//            return ExperimentTaskStatus.FAILED;
            throw new SshSetupException("SSHClient could not connect after "+connectTries+" tries.", lastConnectEx);
        }

        try {
            //we must use the distributed ssh keys to login: in speaksfor case, there is no other way to login
            ssh.authPublickey(loginService.getUsername(), new KeyPairWrapper(loginKeyPair));
        } catch (UserAuthException e) {
            updateMessage("Could not authenticate: " + e.getMessage());
            LOG.error("Authentication failed on {}: {}", node.getClientId(), e.getMessage(), e);
//            return ExperimentTaskStatus.FAILED;
            throw new SshSetupException("Authentication failed on "+node.getClientId(), e);
        } catch (TransportException e) {
            updateMessage("Error while authenticating: " + e.getMessage());
            LOG.error("Error while authenticating on {}: {}", node.getClientId(), e.getMessage(), e);
//            return ExperimentTaskStatus.FAILED;
            throw new SshSetupException("Error while authenticating on "+node.getClientId(), e);
        }

        ssh.getConnection().getKeepAlive().setKeepAliveInterval(30);

        return ssh;
    }

}
